; Copyright (c) 2021, rrc-open-firmware FM10K-Documentation Contributors
; 
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are met:
; 
;     * Redistributions of source code must retain the above copyright notice,
;       this list of conditions and the following disclaimer.
;     * Redistributions in binary form must reproduce the above copyright
;       notice, this list of conditions and the following disclaimer in the
;       documentation and/or other materials provided with the distribution.
;     * Neither the name of the copyright holder nor the names of its contributors
;       may be used to endorse or promote products derived from this software
;       without specific prior written permission.
; 
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
; DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
; CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
; OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.



.constant tmp_PCIE_ADDRESS_OFFSET 0x100000

entrypoint:
    CALL @init_startClocks
    FASTCALL @init_setClockSource, bootCfg_systimeClockSource
    FASTCALL @init_setPcieSplitMode, bootCfg_pep0_mode, bootCfg_pep2_mode, bootCfg_pep4_mode, bootCfg_pep6_mode
    SET DEVICE_CFG, DEVICE_CFG.PCIeEnable, DEVICE_CFG.PCIeEnable; Set PCIeEnable[0-8] to enabled
    CALL @init_markClockStable
    CALL @init_markSbusBooted
    FASTCALL @init_memoryRepairConfig, bootCfg_skipMemRepair
    CALL @init_bistStateCheck
    MOV R0, bootCfg_skipPcieInitialization
    BEQ R0, 1, 1, @.skipPcieInit
        SET SOFT_RESET, 0, 0x6 ; De-assert EPLReset = 0, SwitchReset = 0
        WAIT 10 ; Wait 100ns, minimum
        FASTCALL @init_setPciePEPEnabled, bootCfg_pep0_enable, bootCfg_pep1_enable, bootCfg_pep2_enable, bootCfg_pep3_enable, bootCfg_pep4_enable, bootCfg_pep5_enable, bootCfg_pep6_enable, bootCfg_pep7_enable, bootCfg_pep8_enable
        FASTCALL @init_PCIE_setClockOutDiv, 10

        CALL @lock_PCIE_SBUS_take ; Try to take lock for ourselves
        BNE RRET, 1, 0x1, @.skipPcieInit ; We failed to take the lock

        CALL @init_PCIE_MASTER_SPICO
        BNE RRET, 1, 0x1, @.skipPcieInit ; Failed to init properly

        CALL @init_ALL_PCIE_SERDES_SPICO
        BNE RRET, 1, 0x1, @.skipPcieInit ; Failed to init properly

        ; Restart PEPs
        MOV api_BSM_STATUS, api_BSM_STATUS.PCIe_Warm_Reset_Deassert_Starting
        FASTCALL @init_setPciePEPEnabled, bootCfg_pep0_enable, bootCfg_pep1_enable, bootCfg_pep2_enable, bootCfg_pep3_enable, bootCfg_pep4_enable, bootCfg_pep5_enable, bootCfg_pep6_enable, bootCfg_pep7_enable, bootCfg_pep8_enable
        WAIT 5 ; Wait after enabling PEPs to initialize
        MOV api_BSM_STATUS, api_BSM_STATUS.PCIe_Warm_Reset_Deassert_Completed
        MOV api_PCIE_DE_WARM_RESET_STATUS, api_BSM_STATUS

        FASTCALL @init_PCIE_setClockOutDiv, 6

        CALL @init_PCIE_EACH_SERDES_SPICO
        BNE RRET, 1, 0x1, @.skipPcieInit ; Failed to init properly

        CALL @init_PCIE_SERDES_PCS

        CALL @lock_PCIE_SBUS_release
        SET platform_PCIE_SBUS_LOCK_HOLD, 0x1, 0x1

        MOV R0, bootCfg_GPIO_PIN14_DRIVE
        BNE R0, 0x1, 0x1, @.skipGPIO_PIN14_DRIVE
            SET GPIO_DATA, 0x00000000, 1 <14       ; Drive PIN 14
            SET GPIO_CFG, 0x10001 <14, 0x10001 <14 ; Set PIN 14 to Dir = Output, OpenDrain = NormalOutput
        .skipGPIO_PIN14_DRIVE:


        ; Setup interrupt handler configuration
        MOV R0, @interrupt_handler
        OR R0, R0, bootCfg_spiTransferMode <6 <24 ; bootCfg_spiTransferMode << 6 << 24
        OR R0, R0, bootCfg_spiTransferSpeed <3 <24 ; bootCfg_spiTransferSpeed << 3 << 24
        MOV BSM_ARGS, R0 ; Configure base address, mode and speed for interrupt handler (for PCIE_IP.VPD_Request, or When there is a global interrupt directed to the BSM (i.e. |(GLOBAL_INTERRUPT_DETECT & ~INTERRUPT_MASK_BSM) is true) and BSM_CTRL.Command is set to NONE.)

        ; Setup Interrupt from PEP Mask to Boot State Machine configuration
        MOV R1, BSM_ADDR_OFFSET_1 ; Save any old offset
        MOV BSM_ADDR_OFFSET_1, 0x000000 ; PCIe base address for PEP

        MOV R0, 0x000007ff ; PCIE_IB settings, All enabled
        SET R0, 0, PCIE_IB.VPD_Request ; Mask PCIE_IP.VPD_Request = 0
        SET R0, 0, PCIE_IB.OutOfReset ; Mask PCIE_IP.OutOfReset = 0
        SET R0, 0, PCIE_IB.DataPathReset ; Mask PCIE_IP.DataPathReset = 0
        SET R0, 0, PCIE_IB.PFLR ; Mask PCIE_IP.PFLR = 0

        ; TODO: move this to pcie.asm and make it a function with looping
        BNE DEVICE_CFG, DEVICE_CFG.PCIeEnable_0, DEVICE_CFG.PCIeEnable_0, @.skip_PEP0_masking ; Is PCIeEnable != 1
            SET SOFT_RESET SOFT_RESET.PCIeActive_0, SOFT_RESET.PCIeActive_0 ; Set PEP0 PCIeActive = 1
            MOV~10 PCIE_IB, R0 ; Copy interrupt settings
            SET SOFT_RESET 0, SOFT_RESET.PCIeActive_0 ; Set PEP0 PCIeActive = 0
        .skip_PEP0_masking:
        ADD BSM_ADDR_OFFSET_1, BSM_ADDR_OFFSET_1, tmp_PCIE_ADDRESS_OFFSET

        BNE DEVICE_CFG, DEVICE_CFG.PCIeEnable_1, DEVICE_CFG.PCIeEnable_1, @.skip_PEP1_masking ; Is PCIeEnable != 1
            SET SOFT_RESET SOFT_RESET.PCIeActive_1, SOFT_RESET.PCIeActive_1 ; Set PEP1 PCIeActive = 1
            MOV~10 PCIE_IB, R0 ; Copy interrupt settings
            SET SOFT_RESET 0, SOFT_RESET.PCIeActive_1 ; Set PEP1 PCIeActive = 0
        .skip_PEP1_masking:
        ADD BSM_ADDR_OFFSET_1, BSM_ADDR_OFFSET_1, tmp_PCIE_ADDRESS_OFFSET

        BNE DEVICE_CFG, DEVICE_CFG.PCIeEnable_2, DEVICE_CFG.PCIeEnable_2, @.skip_PEP2_masking ; Is PCIeEnable != 1
            SET SOFT_RESET SOFT_RESET.PCIeActive_2, SOFT_RESET.PCIeActive_2 ; Set PEP2 PCIeActive = 1
            MOV~10 PCIE_IB, R0 ; Copy interrupt settings
            SET SOFT_RESET 0, SOFT_RESET.PCIeActive_2 ; Set PEP2 PCIeActive = 0
        .skip_PEP2_masking:
        ADD BSM_ADDR_OFFSET_1, BSM_ADDR_OFFSET_1, tmp_PCIE_ADDRESS_OFFSET

        BNE DEVICE_CFG, DEVICE_CFG.PCIeEnable_3, DEVICE_CFG.PCIeEnable_3, @.skip_PEP3_masking ; Is PCIeEnable != 1
            SET SOFT_RESET SOFT_RESET.PCIeActive_3, SOFT_RESET.PCIeActive_3 ; Set PEP3 PCIeActive = 1
            MOV~10 PCIE_IB, R0 ; Copy interrupt settings
            SET SOFT_RESET 0, SOFT_RESET.PCIeActive_3 ; Set PEP3 PCIeActive = 0
        .skip_PEP3_masking:
        ADD BSM_ADDR_OFFSET_1, BSM_ADDR_OFFSET_1, tmp_PCIE_ADDRESS_OFFSET

        BNE DEVICE_CFG, DEVICE_CFG.PCIeEnable_4, DEVICE_CFG.PCIeEnable_4, @.skip_PEP4_masking ; Is PCIeEnable != 1
            SET SOFT_RESET SOFT_RESET.PCIeActive_4, SOFT_RESET.PCIeActive_4 ; Set PEP4 PCIeActive = 1
            MOV~10 PCIE_IB, R0 ; Copy interrupt settings
            SET SOFT_RESET 0, SOFT_RESET.PCIeActive_4 ; Set PEP4 PCIeActive = 0
        .skip_PEP4_masking:
        ADD BSM_ADDR_OFFSET_1, BSM_ADDR_OFFSET_1, tmp_PCIE_ADDRESS_OFFSET

        BNE DEVICE_CFG, DEVICE_CFG.PCIeEnable_5, DEVICE_CFG.PCIeEnable_5, @.skip_PEP5_masking ; Is PCIeEnable != 1
            SET SOFT_RESET SOFT_RESET.PCIeActive_5, SOFT_RESET.PCIeActive_5 ; Set PEP5 PCIeActive = 1
            MOV~10 PCIE_IB, R0 ; Copy interrupt settings
            SET SOFT_RESET 0, SOFT_RESET.PCIeActive_5 ; Set PEP5 PCIeActive = 0
        .skip_PEP5_masking:
        ADD BSM_ADDR_OFFSET_1, BSM_ADDR_OFFSET_1, tmp_PCIE_ADDRESS_OFFSET

        BNE DEVICE_CFG, DEVICE_CFG.PCIeEnable_6, DEVICE_CFG.PCIeEnable_6, @.skip_PEP6_masking ; Is PCIeEnable != 1
            SET SOFT_RESET SOFT_RESET.PCIeActive_6, SOFT_RESET.PCIeActive_6 ; Set PEP6 PCIeActive = 1
            MOV~10 PCIE_IB, R0 ; Copy interrupt settings
            SET SOFT_RESET 0, SOFT_RESET.PCIeActive_6 ; Set PEP6 PCIeActive = 0
        .skip_PEP6_masking:
        ADD BSM_ADDR_OFFSET_1, BSM_ADDR_OFFSET_1, tmp_PCIE_ADDRESS_OFFSET

        BNE DEVICE_CFG, DEVICE_CFG.PCIeEnable_7, DEVICE_CFG.PCIeEnable_7, @.skip_PEP7_masking ; Is PCIeEnable != 1
            SET SOFT_RESET SOFT_RESET.PCIeActive_7, SOFT_RESET.PCIeActive_7 ; Set PEP7 PCIeActive = 1
            MOV~10 PCIE_IB, R0 ; Copy interrupt settings
            SET SOFT_RESET 0, SOFT_RESET.PCIeActive_7 ; Set PEP7 PCIeActive = 0
        .skip_PEP7_masking:
        ADD BSM_ADDR_OFFSET_1, BSM_ADDR_OFFSET_1, tmp_PCIE_ADDRESS_OFFSET

        BNE DEVICE_CFG, DEVICE_CFG.PCIeEnable_8, DEVICE_CFG.PCIeEnable_8, @.skip_PEP8_masking ; Is PCIeEnable != 1
            SET SOFT_RESET SOFT_RESET.PCIeActive_8, SOFT_RESET.PCIeActive_8 ; Set PEP8 PCIeActive = 1
            MOV~10 PCIE_IB, R0 ; Copy interrupt settings
            SET SOFT_RESET 0, SOFT_RESET.PCIeActive_8 ; Set PEP8 PCIeActive = 0
        .skip_PEP8_masking:


        MOV BSM_ADDR_OFFSET_1, R1 ; Restore old offset

        CALL @init_finishClockConfig

        SET INTERRUPT_MASK_BSM_0, 0, INTERRUPTS_0.PCIE ; Set PCIE_BSM[0-8] = 0
    .skipPcieInit:
    BEQ platform_PCIE_SBUS_LOCK_HOLD, 1, 0x1, @.noReleaseLock
        CALL @lock_PCIE_SBUS_release
    .noReleaseLock:
    CALL @config_load_bootCfg_customMac
    MOV api_EEPROM_IMAGE_VERSION, platform_firmware_versionNumber    ; Sets the version on memory where a switch manager can read it
    RET

init_markSbusBooted:
    MOV api_BSM_STATUS, api_BSM_STATUS.EPL_PCIE_SBUS_Controller_Reset_Starting
    MOV SBUS_EPL_CFG, 0 ; Set SBUS_ControllerReset = 0 (Not in reset) (RomEnable = 0, which is default)
    MOV SBUS_PCIE_CFG, 0 ; Set SBUS_ControllerReset = 0 (Not in reset) (RomEnable = 0, which is default)
    MOV api_BSM_STATUS, api_BSM_STATUS.EPL_PCIE_SBUS_Controller_Reset_Complete
    MOV api_SBUS_RESET_STATUS, api_BSM_STATUS
    RET

; FASTCALL void(bool skipMemRepair)
init_memoryRepairConfig:
    MOV api_BSM_STATUS, api_BSM_STATUS.Memory_Repair_Starting ; apparently api_MEMORY_REPAIR_STATUS is not set here
    BEQ P0, 0, 0x1, @.noSkipRepair
        MOV api_BSM_STATUS, api_BSM_STATUS.Memory_Repair_Skipped_Config
        MOV api_MEMORY_REPAIR_STATUS, api_BSM_STATUS
        RET
    .noSkipRepair:
    MOV REI_CTRL, 0x9; Set RAM EFUSE interface, Reset = 1, Mode = 0, Run = 0, AutoLoadEnable = 1
    MOV REI_CTRL, 0x8; Take it out of reset. Set RAM EFUSE interface, Reset = 0, Mode = 0, Run = 0, AutoLoadEnable = 1
    MOV api_BSM_STATUS, api_BSM_STATUS.Memory_Repair_Running ; apparently api_MEMORY_REPAIR_STATUS is not set here
    MOV R0, 0x3 ; Number of total tries
    .checkMemTest:
    SUB R0, R0, 1 ; Decrease total tries
    PNE REI_STAT, 1, 0x3, 16000, 10, @.failMemCheck ; Check if ReiDonePass = 1, and ReiDoneFail = 0
        MOV api_BSM_STATUS, api_BSM_STATUS.Memory_Repair_Completed_Pass ; Success!
        MOV api_MEMORY_REPAIR_STATUS, api_BSM_STATUS
        WAIT 10
        RET
    .failMemCheck:
    BEQ REI_STAT, 0x2, 0x2, @.failMem ; Check if ReiDoneFail = 1, and ReiDonePass = 0
        MOV api_BSM_STATUS, api_BSM_STATUS.Memory_Repair_Not_Completed ; Retry
        MOV api_MEMORY_REPAIR_STATUS, api_BSM_STATUS
        BNE R0, 0, 0x3, @.checkMemTest ; Jump if total tries != 0
    .failMem: ; Failed or took too many tries
    MOV api_BSM_STATUS, api_BSM_STATUS.Memory_Repair_Completed_Fail
    MOV api_MEMORY_REPAIR_STATUS, api_BSM_STATUS
    END     ; Program Execution finishes after memory check failed


init_bistStateCheck:
    MOV api_BSM_STATUS, api_BSM_STATUS.Memory_Init_Starting
    WRITE BIST_CTRL_0, 0x00000000, 0x0000ffff ; 64-bit write, regular BIST, BistRun_PCIE[0-8] = 1, BistRun_EPL = 1, BistRun_FABRIC = 1, BistRun_TUNNEL = 1, BistRun_BSM = 1, BistRun_CRM = 1, BistRun_FIBM = 1, BistRun_SBM = 1
    WRITE BIST_CTRL_0, 0x0000ffff, 0x0000ffff ; 64-bit write, BIST memory init, BistRun_PCIE[0-8] = 1, BistRun_EPL = 1, BistRun_FABRIC = 1, BistRun_TUNNEL = 1, BistRun_BSM = 1, BistRun_CRM = 1, BistRun_FIBM = 1, BistRun_SBM = 1
    MOV api_BSM_STATUS, api_BSM_STATUS.Memory_Init_Running
    WAIT 80000
    WRITE BIST_CTRL_0, 0x00000000, 0x0000ffff ; 64-bit write, regular BIST, BistRun_PCIE[0-8] = 1, BistRun_EPL = 1, BistRun_FABRIC = 1, BistRun_TUNNEL = 1, BistRun_BSM = 1, BistRun_CRM = 1, BistRun_FIBM = 1, BistRun_SBM = 1
    WRITE BIST_CTRL_0, 0x00000000, 0x00000000 ; 64-bit write, regular BIST, set all BistRun_* = 0
    MOV api_BSM_STATUS, api_BSM_STATUS.Memory_Init_Stopped
    MOV api_MEMORY_INIT_STATUS, api_BSM_STATUS
    RET