;; Copyright (c) 2021, rrc-open-firmware FM10K-Documentation Contributors
;;
;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions are met:
;;
;;     * Redistributions of source code must retain the above copyright notice,
;;       this list of conditions and the following disclaimer.
;;     * Redistributions in binary form must reproduce the above copyright
;;       notice, this list of conditions and the following disclaimer in the
;;       documentation and/or other materials provided with the distribution.
;;     * Neither the name of the copyright holder nor the names of its contributors
;;       may be used to endorse or promote products derived from this software
;;       without specific prior written permission.
;;
;; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
;; AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
;; IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;; DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
;; DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
;; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
;; CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
;; OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
;; OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

interrupt_handler:
    ; TODO: what happens if we receive an interrupt while on another BSM method? can it happen?
    MOV api_BSM_STATUS, api_BSM_STATUS.ISR_Starting
    WRITE rrcc_STACK_POINTER, rrcc_STACK_START                  ; Reset Stack pointer to its starting position
    ;;WRITE R0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0    ; Zero R0 - RF
    INIT R0, 16 ; Zero R0 - RF
    WRITE api_RE_RESET_MASK_STATUS_1, 0x00000000, 0x00000000    ; 64-bit write

    .takeSoftResetLock: CALL @lock_soft_reset_take
    BNE RRET, 1, 0x1, @.takeSoftResetLock                       ; Failed to take lock, keep retrying in busy loop

    CALL @interrupt_handle_each_source
    ; TODO this

    END


lock_soft_reset_take:
    PNE api_SOFT_RESET_LOCK_STATE, 0, 0x3, 20000, 50, @.failTakeLock
        SET api_SOFT_RESET_LOCK_STATE, platform_SOFT_RESET_LOCK_NVM, 0x3
        BNE api_SOFT_RESET_LOCK_STATE, platform_SOFT_RESET_LOCK_NVM, 0x3, @.retryLock
            RET 1 ; Success
        .retryLock:
        PNE api_SOFT_RESET_LOCK_STATE, 0, 0x3, 20000, 50, @.failTakeLock
        SET api_SOFT_RESET_LOCK_STATE, platform_SOFT_RESET_LOCK_NVM, 0x3
        BNE api_SOFT_RESET_LOCK_STATE, platform_SOFT_RESET_LOCK_NVM, 0x3, @.failTakeLock
            RET 1 ; Success
    .failTakeLock:
    SET api_SW_LOCK_ERR_STATUS, 0x00000100, 0x00000100
    BEQ api_SW_LOCK_ERR_STATUS, 0xfe00, 0xfe00, @.returnFail
        ADD api_SW_LOCK_ERR_STATUS, api_SW_LOCK_ERR_STATUS, 0x200
        RET 0 ; Failure
    .returnFail: SET api_SW_LOCK_ERR_STATUS, 0x0000, 0xfe00
    ADD api_SW_LOCK_ERR_STATUS, api_SW_LOCK_ERR_STATUS, 0x200
    RET 0 ; Failure

lock_soft_reset_release:
    BNE api_SOFT_RESET_LOCK_STATE, platform_SOFT_RESET_LOCK_NVM, 0x3, @.return  ; If we hold the lock
        SET api_SOFT_RESET_LOCK_STATE, platform_SOFT_RESET_LOCK_FREE, 0x3       ; Release lock
    .return:
    RET

interrupt_handle_each_source:
    BNE GLOBAL_INTERRUPT_DETECT_0, INTERRUPTS_0.PCIE_BSM_0, INTERRUPTS_0.PCIE_BSM_0, @.skipPep0 ; Need to handle PEP 0
        MOV api_BSM_STATUS, api_BSM_STATUS.Interrupt_Host_Y
        CALL @interrupt_set_clocks_once_PEP0
    .skipPep0:
    ; TODO complete this
    RET

interrupt_set_clocks_once_PEP0:
    BEQ CHIP_VERSION, 0x00, CHIP_VERSION.Version, @.skipReset
        BEQ PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_0, @.skipReset ; If Mode for PEP group = PCIE_REFCLK
            BNE PCIE_CLK_STAT, 0, PCIE_CLK_STAT.RefclkSel_0, @.skipReset ; If RefclkSel != 0
                AND platform_PEP0_PCIE_CLK_CTRL, PCIE_CLK_CTRL, PCIE_CLK_CTRL.Mode_0
                SET SOFT_RESET, SOFT_RESET.PCIeReset_0, SOFT_RESET.PCIeReset_0 ; Set PCIeReset for PEP = 1
                SET PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_0; Set Mode for PEP group = PCIE_REFCLK
                SET SOFT_RESET, 0, SOFT_RESET.PCIeReset_0 ; Set PCIeReset for PEP = 0
    .skipReset: SET SOFT_RESET, SOFT_RESET.PCIeActive_0, SOFT_RESET.PCIeActive_0 ; Set PCIeActive for PEP = 1
    RET

interrupt_set_clocks_once_PEP1:
    BEQ CHIP_VERSION, 0x00, CHIP_VERSION.Version, @.skipReset
        BEQ PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_0, @.skipReset ; If Mode for PEP group = PCIE_REFCLK
            BNE PCIE_CLK_STAT, 0, PCIE_CLK_STAT.RefclkSel_0, @.skipReset ; If RefclkSel != 0
                AND platform_PEP1_PCIE_CLK_CTRL, PCIE_CLK_CTRL, PCIE_CLK_CTRL.Mode_0
                SET SOFT_RESET, SOFT_RESET.PCIeReset_1, SOFT_RESET.PCIeReset_1 ; Set PCIeReset for PEP = 1
                SET PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_0; Set Mode for PEP group = PCIE_REFCLK
                SET SOFT_RESET, 0, SOFT_RESET.PCIeReset_1 ; Set PCIeReset for PEP = 0
    .skipReset: SET SOFT_RESET, SOFT_RESET.PCIeActive_1, SOFT_RESET.PCIeActive_1 ; Set PCIeActive for PEP = 1
    RET

interrupt_set_clocks_once_PEP2:
    BEQ CHIP_VERSION, 0x00, CHIP_VERSION.Version, @.skipReset
        BEQ PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_2, @.skipReset ; If Mode for PEP group = PCIE_REFCLK
            BNE PCIE_CLK_STAT, 0, PCIE_CLK_STAT.RefclkSel_2, @.skipReset ; If RefclkSel != 0
                AND platform_PEP2_PCIE_CLK_CTRL, PCIE_CLK_CTRL, PCIE_CLK_CTRL.Mode_2
                SET SOFT_RESET, SOFT_RESET.PCIeReset_2, SOFT_RESET.PCIeReset_2 ; Set PCIeReset for PEP = 1
                SET PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_2; Set Mode for PEP group = PCIE_REFCLK
                SET SOFT_RESET, 0, SOFT_RESET.PCIeReset_2 ; Set PCIeReset for PEP = 0
    .skipReset: SET SOFT_RESET, SOFT_RESET.PCIeActive_2, SOFT_RESET.PCIeActive_2 ; Set PCIeActive for PEP = 1
    RET
    
interrupt_set_clocks_once_PEP3:
    BEQ CHIP_VERSION, 0x00, CHIP_VERSION.Version, @.skipReset
        BEQ PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_2, @.skipReset ; If Mode for PEP group = PCIE_REFCLK
            BNE PCIE_CLK_STAT, 0, PCIE_CLK_STAT.RefclkSel_2, @.skipReset ; If RefclkSel != 0
                AND platform_PEP3_PCIE_CLK_CTRL, PCIE_CLK_CTRL, PCIE_CLK_CTRL.Mode_2
                SET SOFT_RESET, SOFT_RESET.PCIeReset_3, SOFT_RESET.PCIeReset_3 ; Set PCIeReset for PEP = 1
                SET PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_2; Set Mode for PEP group = PCIE_REFCLK
                SET SOFT_RESET, 0, SOFT_RESET.PCIeReset_3 ; Set PCIeReset for PEP = 0
    .skipReset: SET SOFT_RESET, SOFT_RESET.PCIeActive_3, SOFT_RESET.PCIeActive_3 ; Set PCIeActive for PEP = 1
    RET

interrupt_set_clocks_once_PEP4:
    BEQ CHIP_VERSION, 0x00, CHIP_VERSION.Version, @.skipReset
        BEQ PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_4, @.skipReset ; If Mode for PEP group = PCIE_REFCLK
            BNE PCIE_CLK_STAT, 0, PCIE_CLK_STAT.RefclkSel_4, @.skipReset ; If RefclkSel != 0
                AND platform_PEP4_PCIE_CLK_CTRL, PCIE_CLK_CTRL, PCIE_CLK_CTRL.Mode_4
                SET SOFT_RESET, SOFT_RESET.PCIeReset_4, SOFT_RESET.PCIeReset_4 ; Set PCIeReset for PEP = 1
                SET PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_4; Set Mode for PEP group = PCIE_REFCLK
                SET SOFT_RESET, 0, SOFT_RESET.PCIeReset_4 ; Set PCIeReset for PEP = 0
    .skipReset: SET SOFT_RESET, SOFT_RESET.PCIeActive_4, SOFT_RESET.PCIeActive_4 ; Set PCIeActive for PEP = 1
    RET

interrupt_set_clocks_once_PEP5:
    BEQ CHIP_VERSION, 0x00, CHIP_VERSION.Version, @.skipReset
        BEQ PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_4, @.skipReset ; If Mode for PEP group = PCIE_REFCLK
            BNE PCIE_CLK_STAT, 0, PCIE_CLK_STAT.RefclkSel_4, @.skipReset ; If RefclkSel != 0
                AND platform_PEP5_PCIE_CLK_CTRL, PCIE_CLK_CTRL, PCIE_CLK_CTRL.Mode_4
                SET SOFT_RESET, SOFT_RESET.PCIeReset_5, SOFT_RESET.PCIeReset_5 ; Set PCIeReset for PEP = 1
                SET PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_4; Set Mode for PEP group = PCIE_REFCLK
                SET SOFT_RESET, 0, SOFT_RESET.PCIeReset_5 ; Set PCIeReset for PEP = 0
    .skipReset: SET SOFT_RESET, SOFT_RESET.PCIeActive_5, SOFT_RESET.PCIeActive_5 ; Set PCIeActive for PEP = 1
    RET
    
interrupt_set_clocks_once_PEP6:
    BEQ CHIP_VERSION, 0x00, CHIP_VERSION.Version, @.skipReset
        BEQ PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_6, @.skipReset ; If Mode for PEP group = PCIE_REFCLK
            BNE PCIE_CLK_STAT, 0, PCIE_CLK_STAT.RefclkSel_6, @.skipReset ; If RefclkSel != 0
                AND platform_PEP6_PCIE_CLK_CTRL, PCIE_CLK_CTRL, PCIE_CLK_CTRL.Mode_6
                SET SOFT_RESET, SOFT_RESET.PCIeReset_6, SOFT_RESET.PCIeReset_6 ; Set PCIeReset for PEP = 1
                SET PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_6; Set Mode for PEP group = PCIE_REFCLK
                SET SOFT_RESET, 0, SOFT_RESET.PCIeReset_6 ; Set PCIeReset for PEP = 0
    .skipReset: SET SOFT_RESET, SOFT_RESET.PCIeActive_6, SOFT_RESET.PCIeActive_6 ; Set PCIeActive for PEP = 1
    RET
    
interrupt_set_clocks_once_PEP7:
    BEQ CHIP_VERSION, 0x00, CHIP_VERSION.Version, @.skipReset
        BEQ PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_6, @.skipReset ; If Mode for PEP group = PCIE_REFCLK
            BNE PCIE_CLK_STAT, 0, PCIE_CLK_STAT.RefclkSel_6, @.skipReset ; If RefclkSel != 0
                AND platform_PEP7_PCIE_CLK_CTRL, PCIE_CLK_CTRL, PCIE_CLK_CTRL.Mode_6
                SET SOFT_RESET, SOFT_RESET.PCIeReset_7, SOFT_RESET.PCIeReset_7 ; Set PCIeReset for PEP = 1
                SET PCIE_CLK_CTRL, 0, PCIE_CLK_CTRL.Mode_6; Set Mode for PEP group = PCIE_REFCLK
                SET SOFT_RESET, 0, SOFT_RESET.PCIeReset_7 ; Set PCIeReset for PEP = 0
    .skipReset: SET SOFT_RESET, SOFT_RESET.PCIeActive_7, SOFT_RESET.PCIeActive_7 ; Set PCIeActive for PEP = 1
    RET